#ifndef MTL_MATRIX_STREAM_H
#define MTL_MATRIX_STREAM_H

//
// Forward Declarations for the Matrix Streams
//

namespace mtl {

template <class T>
class matrix_market_stream;

template <class T>
class harwell_boeing_stream;

} /* namespace mtl */

#endif

